@@ -133,6 +133,12 @@ typedef struct
   guint use_gvariant : 1;
 } JsonrpcClientPrivate;
 
+typedef struct
+{
+  GHashTable *invocations;
+  GError *error;
+} PanicData;
+
 G_DEFINE_TYPE_WITH_PRIVATE (JsonrpcClient, jsonrpc_client, G_TYPE_OBJECT)
 
 enum {
@@ -212,6 +218,28 @@ is_jsonrpc_call (GVariantDict *dict)
           g_variant_dict_contains (dict, "params"));
 }
 
+static gboolean
+error_invocations_from_idle (gpointer data)
+{
+  PanicData *pd = data;
+  GHashTableIter iter;
+  GTask *task;
+
+  g_assert (pd != NULL);
+  g_assert (pd->invocations != NULL);
+  g_assert (pd->error != NULL);
+
+  g_hash_table_iter_init (&iter, pd->invocations);
+  while (g_hash_table_iter_next (&iter, NULL, (gpointer *)&task))
+    g_task_return_error (task, g_error_copy (pd->error));
+
+  g_clear_pointer (&pd->invocations, g_hash_table_unref);
+  g_clear_pointer (&pd->error, g_error_free);
+  g_slice_free (PanicData, pd);
+
+  return G_SOURCE_REMOVE;
+}
+
 /*
  * jsonrpc_client_panic:
  *
@@ -223,19 +251,27 @@ jsonrpc_client_panic (JsonrpcClient *self,
                       const GError  *error)
 {
   JsonrpcClientPrivate *priv = jsonrpc_client_get_instance_private (self);
-  g_autoptr(GHashTable) invocations = NULL;
-  GHashTableIter iter;
-  GTask *task;
+  g_autoptr(JsonrpcClient) hold = NULL;
+  PanicData *pd;
 
   g_assert (JSONRPC_IS_CLIENT (self));
   g_assert (error != NULL);
 
-  g_object_ref (self);
+  hold = g_object_ref (self);
 
   priv->failed = TRUE;
 
-  /* Steal the tasks so that we don't have to worry about reentry. */
-  invocations = g_steal_pointer (&priv->invocations);
+  /*
+   * Defer the completion of all tasks (with errors) until we've made it
+   * back to the main loop. Otherwise, we get into difficult to determine
+   * re-entrancy cases.
+   */
+  pd = g_slice_new0 (PanicData);
+  pd->invocations = g_steal_pointer (&priv->invocations);
+  pd->error = g_error_copy (error);
+  g_idle_add_full (G_PRIORITY_LOW, error_invocations_from_idle, pd, NULL);
+
+  /* Keep a hashtable around for code that expects a pointer there */
   priv->invocations = g_hash_table_new_full (NULL, NULL, NULL, g_object_unref);
 
   /* Now close the connection */
@@ -248,17 +284,7 @@ jsonrpc_client_panic (JsonrpcClient *self,
   g_clear_object (&priv->input_stream);
   g_clear_object (&priv->output_stream);
 
-  /*
-   * Now notify all of the in-flight invocations that they failed due
-   * to an unrecoverable error.
-   */
-  g_hash_table_iter_init (&iter, invocations);
-  while (g_hash_table_iter_next (&iter, NULL, (gpointer *)&task))
-    g_task_return_error (task, g_error_copy (error));
-
-  g_signal_emit (self, signals [FAILED], 0);
-
-  g_object_unref (self);
+  g_signal_emit (hold, signals [FAILED], 0);
 }
 
 /*
@@ -875,12 +901,12 @@ jsonrpc_client_call (JsonrpcClient  *self,
  *
  * Asynchronously calls @method with @params on the remote peer.
  *
- * This function takes ownership of @params.
- *
  * Upon completion or failure, @callback is executed and it should
  * call jsonrpc_client_call_finish() to complete the request and release
  * any memory held.
  *
+ * If @params is floating, the floating reference is consumed.
+ *
  * Since: 3.26
  */
 void
@@ -1008,13 +1034,13 @@ jsonrpc_client_send_notification_write_cb (GObject      *object,
  * jsonrpc_client_send_notification:
  * @self: A #JsonrpcClient
  * @method: the name of the method to call
- * @params: (transfer full) (nullable): A #JsonNode of parameters or %NULL
+ * @params: (transfer none) (nullable): A #GVariant of parameters or %NULL
  * @cancellable: (nullable): A #GCancellable or %NULL
  *
  * Synchronously calls @method with @params on the remote peer.
  * This function will not wait or expect a reply from the peer.
  *
- * This function takes ownership of @params.
+ * If @params is floating then the reference is consumed.
  *
  * Returns: %TRUE on success; otherwise %FALSE and @error is set.
  *
@@ -1285,6 +1311,24 @@ jsonrpc_client_reply_error_cb (GObject      *object,
     g_task_return_boolean (task, TRUE);
 }
 
+/**
+ * jsonrpc_client_reply_error_async:
+ * @self: a #JsonrpcClient
+ * @id: (transfer none): a #GVariant containing the call id
+ * @code: the error code
+ * @message: (nullable): an optional error message
+ * @cancellable: (nullable): a #GCancellable, or %NULL
+ * @callback: (nullable): a #GAsyncReadyCallback or %NULL
+ * @user_data: closure data for @callback
+ *
+ * Asynchronously replies to the peer, sending a JSON-RPC error message.
+ *
+ * Call jsonrpc_client_reply_error_finish() to get the result of this operation.
+ *
+ * If @id is floating, it's floating reference is consumed.
+ *
+ * Since: 3.28
+ */
 void
 jsonrpc_client_reply_error_async (JsonrpcClient       *self,
                                   GVariant            *id,
@@ -1350,11 +1394,16 @@ jsonrpc_client_reply_error_finish (JsonrpcClient  *self,
 
 /**
  * jsonrpc_client_reply:
+ * @self: a #JsonrpcClient
  * @id: (transfer none) (not nullable): the id of the message to reply
  * result: (transfer none) (nullable): the return value or %NULL
+ * @cancellable: (nullable): a #GCancellable, or %NULL
+ * @error: a #GError, or %NULL
  *
  * Synchronous variant of jsonrpc_client_reply_async().
  *
+ * If @id or @result are floating, there floating references are consumed.
+ *
  * Since: 3.26
  */
 gboolean
@@ -1411,8 +1460,12 @@ jsonrpc_client_reply_cb (GObject      *object,
 
 /**
  * jsonrcp_client_reply_async:
- * @id: (transfer full) (not nullable): the id of the message to reply
- * result: (transfer full) (nullable): the return value or %NULL
+ * @self: a #JsonrpcClient
+ * @id: (transfer none): the id of the message to reply
+ * @result: (transfer none) (nullable): the return value or %NULL
+ * @cancellable: a #GCancellable, or %NULL
+ * @callback: (nullable): a #GAsyncReadyCallback or %NULL
+ * @user_data: closure data for @callback
  *
  * This function will reply to a method call identified by @id with the
  * result provided. If @result is %NULL, then a null JSON node is returned
@@ -1430,6 +1483,8 @@ jsonrpc_client_reply_cb (GObject      *object,
  * asynchronous message does not indicate that the peer has received
  * the message.
  *
+ * If @id or @result are floating, there floating references are consumed.
+ *
  * Since: 3.26
  */
 void
